# coding: utf8
from __future__ import unicode_literals

from ..punctuation import TOKENIZER_INFIXES
from ..char_classes import LIST_PUNCT, LIST_ELLIPSES, LIST_QUOTES, CURRENCY
from ..char_classes import QUOTES, UNITS, ALPHA, ALPHA_LOWER, ALPHA_UPPER

_suffixes = (LIST_PUNCT + LIST_ELLIPSES + LIST_QUOTES +
             [r'(?<=[0-9])\+',
              # Arabic is written from Right-To-Left
              r'(?<=[0-9])(?:{})'.format(CURRENCY),
              r'(?<=[0-9])(?:{})'.format(UNITS),
              r'(?<=[{au}][{au}])\.'.format(au=ALPHA_UPPER)])

TOKENIZER_SUFFIXES = _suffixes



# coding: utf8

"""  
Tags are a combination of POS and morphological features from a yet
unpublished dataset developed by Schibsted, Nasjonalbiblioteket and LTG. The
data format is .conllu and follows the Universal Dependencies annotation. (There
are some annotation differences compared to this dataset:
https://github.com/UniversalDependencies/UD_Norwegian-Bokmaal 
mainly in the way determiners and pronouns are tagged).
"""

from __future__ import unicode_literals

from ...symbols import POS, PUNCT, ADJ, CONJ, CCONJ, SCONJ, SYM, NUM, DET, ADV, ADP, X, VERB
from ...symbols import NOUN, PROPN, PART, INTJ, SPACE, PRON, AUX

TAG_MAP = {
    'ADJ__Case=Gen|Definite=Def|Degree=Pos|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Degree=Pos|Number=Sing', POS: ADJ},
    'ADJ__Case=Gen|Definite=Def|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Number=Sing', POS: ADJ},
    'ADJ__Case=Gen|Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing', POS: ADJ},
    'ADJ__Case=Gen|Definite=Ind|Degree=Pos|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Degree=Pos|Number=Sing', POS: ADJ},
    'ADJ__Case=Gen|Degree=Cmp': {'morph': 'Case=Gen|Degree=Cmp', POS: ADJ},
    'ADJ__Case=Gen|Degree=Pos|Number=Plur': {'morph': 'Case=Gen|Degree=Pos|Number=Plur', POS: ADJ},
    'ADJ__Definite=Def|Degree=Pos|Gender=Masc|Number=Sing': {'morph': 'Definite=Def|Degree=Pos|Gender=Masc|Number=Sing', POS: ADJ},
    'ADJ__Definite=Def|Degree=Pos|Number=Sing': {'morph': 'Definite=Def|Degree=Pos|Number=Sing', POS: ADJ},
    'ADJ__Definite=Def|Degree=Sup': {'morph': 'Definite=Def|Degree=Sup', POS: ADJ},
    'ADJ__Definite=Def|Number=Sing': {'morph': 'Definite=Def|Number=Sing', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Pos': {'morph': 'Definite=Ind|Degree=Pos', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Pos|Gender=Masc|Number=Sing': {'morph': 'Definite=Ind|Degree=Pos|Gender=Masc|Number=Sing', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Pos|Number=Sing': {'morph': 'Definite=Ind|Degree=Pos|Number=Sing', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Sup': {'morph': 'Definite=Ind|Degree=Sup', POS: ADJ},
    'ADJ__Definite=Ind|Gender=Masc|Number=Sing': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing', POS: ADJ},
    'ADJ__Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing', POS: ADJ},
    'ADJ__Definite=Ind|Number=Sing': {'morph': 'Definite=Ind|Number=Sing', POS: ADJ},
    'ADJ__Degree=Cmp': {'morph': 'Degree=Cmp', POS: ADJ},
    'ADJ__Degree=Pos': {'morph': 'Degree=Pos', POS: ADJ},
    'ADJ__Degree=Pos|Number=Plur': {'morph': 'Degree=Pos|Number=Plur', POS: ADJ},
    'ADJ__Degree=Sup': {'morph': 'Degree=Sup', POS: ADJ},
    'ADJ__Number=Plur': {'morph': 'Number=Plur', POS: ADJ},
    'ADJ__Number=Plur|VerbForm=Part': {'morph': 'Number=Plur|VerbForm=Part', POS: ADJ},
    'ADJ__Number=Sing': {'morph': 'Number=Sing', POS: ADJ},
    'ADJ___': {'morph': '_', POS: ADJ},
    'ADP___': {'morph': '_', POS: ADP},
    'ADV___': {'morph': '_', POS: ADV},
    'AUX__Mood=Imp|VerbForm=Fin': {'morph': 'Mood=Imp|VerbForm=Fin', POS: AUX},
    'AUX__Mood=Ind|Tense=Past|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Past|VerbForm=Fin', POS: AUX},
    'AUX__Mood=Ind|Tense=Pres|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin', POS: AUX},
    'AUX__Mood=Ind|Tense=Pres|VerbForm=Fin|Voice=Pass': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin|Voice=Pass', POS: AUX},
    'AUX__VerbForm=Inf': {'morph': 'VerbForm=Inf', POS: AUX},
    'AUX__VerbForm=Part': {'morph': 'VerbForm=Part', POS: AUX},
    'CONJ___': {'morph': '_', POS: CONJ},
    'DET__Case=Gen|Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'DET__Case=Gen|Gender=Fem|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Gender=Fem|Number=Sing|PronType=Dem', POS: DET},
    'DET__Case=Gen|Gender=Masc|Number=Sing': {'morph': 'Case=Gen|Gender=Masc|Number=Sing', POS: DET},
    'DET__Case=Gen|Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'DET__Case=Gen|Gender=Neut|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Gender=Neut|Number=Sing|PronType=Dem', POS: DET},
    'DET__Case=Gen|Number=Plur': {'morph': 'Case=Gen|Number=Plur', POS: DET},
    'DET__Case=Gen|Number=Plur|PronType=Dem': {'morph': 'Case=Gen|Number=Plur|PronType=Dem', POS: DET},
    'DET__Definite=Def': {'morph': 'Definite=Def', POS: DET},
    'DET__Definite=Def|Number=Sing|PronType=Dem': {'morph': 'Definite=Def|Number=Sing|PronType=Dem', POS: DET},
    'DET__Definite=Def|PronType=Dem': {'morph': 'Definite=Def|PronType=Dem', POS: DET},
    'DET__Definite=Ind|Gender=Fem|Number=Sing': {'morph': 'Definite=Ind|Gender=Fem|Number=Sing', POS: DET},
    'DET__Definite=Ind|Gender=Fem|Number=Sing|PronType=Dem': {'morph': 'Definite=Ind|Gender=Fem|Number=Sing|PronType=Dem', POS: DET},
    'DET__Definite=Ind|Gender=Masc|Number=Sing': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing', POS: DET},
    'DET__Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'DET__Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing', POS: DET},
    'DET__Definite=Ind|Gender=Neut|Number=Sing|PronType=Dem': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing|PronType=Dem', POS: DET},
    'DET__Degree=Pos|Number=Plur': {'morph': 'Degree=Pos|Number=Plur', POS: DET},
    'DET__Gender=Fem|Number=Sing': {'morph': 'Gender=Fem|Number=Sing', POS: DET},
    'DET__Gender=Fem|Number=Sing|Poss=Yes': {'morph': 'Gender=Fem|Number=Sing|Poss=Yes', POS: DET},
    'DET__Gender=Fem|Number=Sing|PronType=Dem': {'morph': 'Gender=Fem|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Fem|Number=Sing|PronType=Int': {'morph': 'Gender=Fem|Number=Sing|PronType=Int', POS: DET},
    'DET__Gender=Masc|Number=Sing': {'morph': 'Gender=Masc|Number=Sing', POS: DET},
    'DET__Gender=Masc|Number=Sing|Poss=Yes': {'morph': 'Gender=Masc|Number=Sing|Poss=Yes', POS: DET},
    'DET__Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Masc|Number=Sing|PronType=Int': {'morph': 'Gender=Masc|Number=Sing|PronType=Int', POS: DET},
    'DET__Gender=Neut|Number=Sing': {'morph': 'Gender=Neut|Number=Sing', POS: DET},
    'DET__Gender=Neut|Number=Sing|Poss=Yes': {'morph': 'Gender=Neut|Number=Sing|Poss=Yes', POS: DET},
    'DET__Gender=Neut|Number=Sing|PronType=Dem': {'morph': 'Gender=Neut|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Neut|Number=Sing|PronType=Int': {'morph': 'Gender=Neut|Number=Sing|PronType=Int', POS: DET},
    'DET__Number=Plur': {'morph': 'Number=Plur', POS: DET},
    'DET__Number=Plur|Poss=Yes': {'morph': 'Number=Plur|Poss=Yes', POS: DET},
    'DET__Number=Plur|Poss=Yes|PronType=Rcp': {'morph': 'Number=Plur|Poss=Yes|PronType=Rcp', POS: DET},
    'DET__Number=Plur|PronType=Dem': {'morph': 'Number=Plur|PronType=Dem', POS: DET},
    'DET__Number=Plur|PronType=Int': {'morph': 'Number=Plur|PronType=Int', POS: DET},
    'DET___': {'morph': '_', POS: DET},
    'INTJ___': {'morph': '_', POS: INTJ},
    'NOUN__Case=Gen': {'morph': 'Case=Gen', POS: NOUN},
    'NOUN__Case=Gen|Definite=Def|Gender=Fem|Number=Plur': {'morph': 'Case=Gen|Definite=Def|Gender=Fem|Number=Plur', POS: NOUN},
    'NOUN__Case=Gen|Definite=Def|Gender=Fem|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Gender=Fem|Number=Sing', POS: NOUN},
    'NOUN__Case=Gen|Definite=Def|Gender=Masc|Number=Plur': {'morph': 'Case=Gen|Definite=Def|Gender=Masc|Number=Plur', POS: NOUN},
    'NOUN__Case=Gen|Definite=Def|Gender=Masc|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Gender=Masc|Number=Sing', POS: NOUN},
    'NOUN__Case=Gen|Definite=Def|Gender=Neut|Number=Plur': {'morph': 'Case=Gen|Definite=Def|Gender=Neut|Number=Plur', POS: NOUN},
    'NOUN__Case=Gen|Definite=Def|Gender=Neut|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Gender=Neut|Number=Sing', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Fem|Number=Plur': {'morph': 'Case=Gen|Definite=Ind|Gender=Fem|Number=Plur', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Fem|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Gender=Fem|Number=Sing', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Masc|Number=Plur': {'morph': 'Case=Gen|Definite=Ind|Gender=Masc|Number=Plur', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Masc|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Gender=Masc|Number=Sing', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Neut|Number=Plur': {'morph': 'Case=Gen|Definite=Ind|Gender=Neut|Number=Plur', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Gender=Neut|Number=Sing', POS: NOUN},
    'NOUN__Case=Gen|Gender=Fem': {'morph': 'Case=Gen|Gender=Fem', POS: NOUN},
    'NOUN__Definite=Def,Ind|Gender=Masc|Number=Plur,Sing': {'morph': 'Definite=Def', POS: NOUN},
    'NOUN__Definite=Def,Ind|Gender=Masc|Number=Sing': {'morph': 'Definite=Def', POS: NOUN},
    'NOUN__Definite=Def,Ind|Gender=Neut|Number=Plur,Sing': {'morph': 'Definite=Def', POS: NOUN},
    'NOUN__Definite=Def|Gender=Fem|Number=Plur': {'morph': 'Definite=Def|Gender=Fem|Number=Plur', POS: NOUN},
    'NOUN__Definite=Def|Gender=Fem|Number=Sing': {'morph': 'Definite=Def|Gender=Fem|Number=Sing', POS: NOUN},
    'NOUN__Definite=Def|Gender=Masc|Number=Plur': {'morph': 'Definite=Def|Gender=Masc|Number=Plur', POS: NOUN},
    'NOUN__Definite=Def|Gender=Masc|Number=Sing': {'morph': 'Definite=Def|Gender=Masc|Number=Sing', POS: NOUN},
    'NOUN__Definite=Def|Gender=Neut|Number=Plur': {'morph': 'Definite=Def|Gender=Neut|Number=Plur', POS: NOUN},
    'NOUN__Definite=Def|Gender=Neut|Number=Sing': {'morph': 'Definite=Def|Gender=Neut|Number=Sing', POS: NOUN},
    'NOUN__Definite=Def|Number=Plur': {'morph': 'Definite=Def|Number=Plur', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Fem|Number=Plur': {'morph': 'Definite=Ind|Gender=Fem|Number=Plur', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Fem|Number=Sing': {'morph': 'Definite=Ind|Gender=Fem|Number=Sing', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Masc': {'morph': 'Definite=Ind|Gender=Masc', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Masc|Number=Plur': {'morph': 'Definite=Ind|Gender=Masc|Number=Plur', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Masc|Number=Sing': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Neut|Number=Plur': {'morph': 'Definite=Ind|Gender=Neut|Number=Plur', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing', POS: NOUN},
    'NOUN__Definite=Ind|Number=Plur': {'morph': 'Definite=Ind|Number=Plur', POS: NOUN},
    'NOUN__Gender=Fem': {'morph': 'Gender=Fem', POS: NOUN},
    'NOUN__Gender=Masc': {'morph': 'Gender=Masc', POS: NOUN},
    'NOUN__Gender=Masc|Number=Sing': {'morph': 'Gender=Masc|Number=Sing', POS: NOUN},
    'NOUN__Gender=Neut': {'morph': 'Gender=Neut', POS: NOUN},
    'NOUN__Number=Plur': {'morph': 'Number=Plur', POS: NOUN},
    'NOUN___': {'morph': '_', POS: NOUN},
    'NUM__Case=Gen|Number=Plur': {'morph': 'Case=Gen|Number=Plur', POS: NUM},
    'NUM__Definite=Def': {'morph': 'Definite=Def', POS: NUM},
    'NUM__Definite=Def|Number=Sing': {'morph': 'Definite=Def|Number=Sing', POS: NUM},
    'NUM__Gender=Fem|Number=Sing': {'morph': 'Gender=Fem|Number=Sing', POS: NUM},
    'NUM__Gender=Masc|Number=Sing': {'morph': 'Gender=Masc|Number=Sing', POS: NUM},
    'NUM__Gender=Neut|Number=Sing': {'morph': 'Gender=Neut|Number=Sing', POS: NUM},
    'NUM__Number=Plur': {'morph': 'Number=Plur', POS: NUM},
    'NUM__Number=Sing': {'morph': 'Number=Sing', POS: NUM},
    'NUM___': {'morph': '_', POS: NUM},
    'PART___': {'morph': '_', POS: PART},
    'PRON__Animacy=Anim|Case=Acc|Gender=Fem|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Anim|Case=Acc|Gender=Fem|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Acc|Gender=Masc|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Anim|Case=Acc|Gender=Masc|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Acc|Number=Plur|Person=1|PronType=Prs': {'morph': 'Animacy=Anim|Case=Acc|Number=Plur|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Acc|Number=Plur|Person=2|PronType=Prs': {'morph': 'Animacy=Anim|Case=Acc|Number=Plur|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Acc|Number=Sing|Person=1|PronType=Prs': {'morph': 'Animacy=Anim|Case=Acc|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Acc|Number=Sing|Person=2|PronType=Prs': {'morph': 'Animacy=Anim|Case=Acc|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Gen,Nom|Number=Sing|PronType=Prs': {'morph': 'Animacy=Anim|Case=Gen', POS: PRON},
    'PRON__Animacy=Anim|Case=Gen|Number=Sing|PronType=Prs': {'morph': 'Animacy=Anim|Case=Gen|Number=Sing|PronType=Prs', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Gender=Fem|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Gender=Fem|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Gender=Masc|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Gender=Masc|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Number=Plur|Person=1|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Number=Plur|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Number=Plur|Person=2|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Number=Plur|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Number=Sing|Person=1|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Number=Sing|Person=2|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Number=Sing|Person=', POS: PRON},
    'PRON__Animacy=Anim|Case=Nom|Number=Sing|PronType=Prs': {'morph': 'Animacy=Anim|Case=Nom|Number=Sing|PronType=Prs', POS: PRON},
    'PRON__Animacy=Anim|Number=Plur|PronType=Rcp': {'morph': 'Animacy=Anim|Number=Plur|PronType=Rcp', POS: PRON},
    'PRON__Animacy=Anim|Number=Sing|PronType=Prs': {'morph': 'Animacy=Anim|Number=Sing|PronType=Prs', POS: PRON},
    'PRON__Animacy=Anim|Poss=Yes|PronType=Int': {'morph': 'Animacy=Anim|Poss=Yes|PronType=Int', POS: PRON},
    'PRON__Animacy=Anim|PronType=Int': {'morph': 'Animacy=Anim|PronType=Int', POS: PRON},
    'PRON__Case=Acc|Number=Plur|Person=3|PronType=Prs': {'morph': 'Case=Acc|Number=Plur|Person=', POS: PRON},
    'PRON__Case=Acc|Reflex=Yes': {'morph': 'Case=Acc|Reflex=Yes', POS: PRON},
    'PRON__Case=Nom|Number=Plur|Person=3|PronType=Prs': {'morph': 'Case=Nom|Number=Plur|Person=', POS: PRON},
    'PRON__Gender=Fem,Masc|Number=Sing|Person=3|PronType=Prs': {'morph': 'Gender=Fem', POS: PRON},
    'PRON__Gender=Neut|Number=Sing|Person=3|PronType=Prs': {'morph': 'Gender=Neut|Number=Sing|Person=', POS: PRON},
    'PRON__Number=Plur|Person=3|PronType=Prs': {'morph': 'Number=Plur|Person=', POS: PRON},
    'PRON__Number=Sing': {'morph': 'Number=Sing', POS: PRON},
    'PRON__PronType=Int': {'morph': 'PronType=Int', POS: PRON},
    'PRON___': {'morph': '_', POS: PRON},
    'PROPN__Case=Gen': {'morph': 'Case=Gen', POS: PROPN},
    'PROPN__Case=Gen|Gender=Fem': {'morph': 'Case=Gen|Gender=Fem', POS: PROPN},
    'PROPN__Case=Gen|Gender=Masc': {'morph': 'Case=Gen|Gender=Masc', POS: PROPN},
    'PROPN__Case=Gen|Gender=Neut': {'morph': 'Case=Gen|Gender=Neut', POS: PROPN},
    'PROPN__Gender=Fem': {'morph': 'Gender=Fem', POS: PROPN},
    'PROPN__Gender=Masc': {'morph': 'Gender=Masc', POS: PROPN},
    'PROPN__Gender=Neut': {'morph': 'Gender=Neut', POS: PROPN},
    'PROPN___': {'morph': '_', POS: PROPN},
    'PUNCT___': {'morph': '_', POS: PUNCT},
    'SCONJ___': {'morph': '_', POS: SCONJ},
    'SYM___': {'morph': '_', POS: SYM},
    'VERB__Definite=Ind|Number=Sing': {'morph': 'Definite=Ind|Number=Sing', POS: VERB},
    'VERB__Mood=Imp|VerbForm=Fin': {'morph': 'Mood=Imp|VerbForm=Fin', POS: VERB},
    'VERB__Mood=Ind|Tense=Past|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Past|VerbForm=Fin', POS: VERB},
    'VERB__Mood=Ind|Tense=Past|VerbForm=Fin|Voice=Pass': {'morph': 'Mood=Ind|Tense=Past|VerbForm=Fin|Voice=Pass', POS: VERB},
    'VERB__Mood=Ind|Tense=Pres|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin', POS: VERB},
    'VERB__Mood=Ind|Tense=Pres|VerbForm=Fin|Voice=Pass': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin|Voice=Pass', POS: VERB},
    'VERB__VerbForm=Inf': {'morph': 'VerbForm=Inf', POS: VERB},
    'VERB__VerbForm=Inf|Voice=Pass': {'morph': 'VerbForm=Inf|Voice=Pass', POS: VERB},
    'VERB__VerbForm=Part': {'morph': 'VerbForm=Part', POS: VERB},
    'VERB___': {'morph': '_', POS: VERB},
    'X___': {'morph': '_', POS: X}
}




# coding: utf8
from __future__ import unicode_literals


PRONOUNS_IRREG = {
    "aucune": ("aucun",),
    "celle-ci": ("celui-ci",),
    "celles-ci": ("celui-ci",),
    "ceux-ci": ("celui-ci",),
    "celle-là": ("celui-là",),
    "celles-là": ("celui-là",),
    "ceux-là": ("celui-là",),
    "celle": ("celui",),
    "celles": ("celui",),
    "ceux": ("celui",),
    "certaines": ("certains",),
    "chacune": ("chacun",),
    "icelle": ("icelui",),
    "icelles": ("icelui",),
    "iceux": ("icelui",),
    "la": ("le",),
    "les": ("le",),
    "laquelle": ("lequel",),
    "lesquelles": ("lequel",),
    "lesquels": ("lequel",),
    "elle-même": ("lui-même",),
    "elles-mêmes": ("lui-même",),
    "eux-mêmes": ("lui-même",),
    "quelle": ("quel",),
    "quelles": ("quel",),
    "quels": ("quel",),
    "quelques-unes": ("quelqu'un",),
    "quelques-uns": ("quelqu'un",),
    "quelque-une": ("quelqu'un",),
    "qu": ("que",),
    "telle": ("tel",),
    "telles": ("tel",),
    "tels": ("tel",),
    "toutes": ("tous",),
}


# coding: utf8
from __future__ import unicode_literals

import functools

from ..errors import Errors


class Underscore(object):
    doc_extensions = {}
    span_extensions = {}
    token_extensions = {}

    def __init__(self, extensions, obj, start=None, end=None):
        object.__setattr__(self, '_extensions', extensions)
        object.__setattr__(self, '_obj', obj)
        # Assumption is that for doc values, _start and _end will both be None
        # Span will set non-None values for _start and _end
        # Token will have _start be non-None, _end be None
        # This lets us key everything into the doc.user_data dictionary,
        # (see _get_key), and lets us use a single Underscore class.
        object.__setattr__(self, '_doc', obj.doc)
        object.__setattr__(self, '_start', start)
        object.__setattr__(self, '_end', end)

    def __getattr__(self, name):
        if name not in self._extensions:
            raise AttributeError(Errors.E046.format(name=name))
        default, method, getter, setter = self._extensions[name]
        if getter is not None:
            return getter(self._obj)
        elif method is not None:
            return functools.partial(method, self._obj)
        else:
            return self._doc.user_data.get(self._get_key(name), default)

    def __setattr__(self, name, value):
        if name not in self._extensions:
            raise AttributeError(Errors.E047.format(name=name))
        default, method, getter, setter = self._extensions[name]
        if setter is not None:
            return setter(self._obj, value)
        else:
            self._doc.user_data[self._get_key(name)] = value

    def set(self, name, value):
        return self.__setattr__(name, value)

    def get(self, name):
        return self.__getattr__(name)

    def has(self, name):
        return name in self._extensions

    def _get_key(self, name):
        return ('._.', name, self._start, self._end)


def get_ext_args(**kwargs):
    """Validate and convert arguments. Reused in Doc, Token and Span."""
    default = kwargs.get('default')
    getter = kwargs.get('getter')
    setter = kwargs.get('setter')
    method = kwargs.get('method')
    if getter is None and setter is not None:
        raise ValueError(Errors.E089)
    valid_opts = ('default' in kwargs, method is not None, getter is not None)
    nr_defined = sum(t is True for t in valid_opts)
    if nr_defined != 1:
        raise ValueError(Errors.E083.format(nr_defined=nr_defined))
    if setter is not None and not hasattr(setter, '__call__'):
        raise ValueError(Errors.E091.format(name='setter', value=repr(setter)))
    if getter is not None and not hasattr(getter, '__call__'):
        raise ValueError(Errors.E091.format(name='getter', value=repr(getter)))
    if method is not None and not hasattr(method, '__call__'):
        raise ValueError(Errors.E091.format(name='method', value=repr(method)))
    return (default, method, getter, setter)
